@@ -1,78 +1,79 @@
-"""
-Make 'span' in tags dict a stack
-maybe do the same for all tags in case of unclosed tags?
-optionally use bs4 to clean up invalid html?
-
-the idea is that there is a method that converts html files into docx
-but also have api methods that let user have more control e.g. so they
-can nest calls to something like 'convert_chunk' in loops
-
-user can pass existing document object as arg
-(if they want to manage rest of document themselves)
-
-How to deal with block level style applied over table elements? e.g. text align
-"""
 import argparse
 import os
 import re
-import base64
-import urllib
-
 from io import BytesIO
 from html.parser import HTMLParser
 
-from bs4 import BeautifulSoup
-
 import docx
+from bs4 import BeautifulSoup
 from docx import Document
-from docx.shared import RGBColor, Inches
 from docx.enum.text import WD_ALIGN_PARAGRAPH
+from docx.enum.table import WD_ALIGN_VERTICAL
 from docx.oxml import OxmlElement
 from docx.oxml.ns import qn
+from docx.shared import RGBColor
 
+from functools import lru_cache
+
+from html4docx import constants
 from html4docx import utils
-
-# values in inches
-INDENT = 0.25
-LIST_INDENT = 0.5
-MAX_INDENT = 5.5 # To stop indents going off the page
-
-# Style to use with tables. By default no style is used.
-DEFAULT_TABLE_STYLE = None
+from html4docx.metadata import Metadata
 
 class HtmlToDocx(HTMLParser):
+    """
+        Class to convert HTML to Docx
+        source: https://docs.python.org/3/library/html.parser.html
+    """
     def __init__(self):
         super().__init__()
-        self.options = {
-            'fix-html': True,
-            'images': True,
-            'tables': True,
-            'styles': True,
-        }
-        self.table_row_selectors = [
-            'table > tr',
-            'table > thead > tr',
-            'table > tbody > tr',
-            'table > tfoot > tr'
-        ]
-        self.table_style = DEFAULT_TABLE_STYLE
+        self.options = dict(constants.DEFAULT_OPTIONS)
+        self.table_row_selectors = constants.DEFAULT_TABLE_ROW_SELECTORS
+        self.table_style = constants.DEFAULT_TABLE_STYLE
 
-    def set_initial_attrs(self, document=None):
+    def set_initial_attrs(self, document = None):
         self.tags = {
             'span': [],
             'list': [],
         }
         self.doc = document if document else Document()
-        self.document = self.doc
         self.bs = self.options['fix-html'] # whether or not to clean with BeautifulSoup
-        self.include_tables = True # TODO add this option back in?
-        self.include_images = self.options['images']
-        self.include_styles = self.options['styles']
         self.paragraph = None
         self.skip = False
         self.skip_tag = None
         self.instances_to_skip = 0
         self.bookmark_id = 0
+        self.in_li = False
+        self.list_restart_counter = 0
+        self.current_ol_num_id = None
+        self._list_num_ids = {}
+
+    @property
+    def metadata(self) -> dict[str, any]:
+        if not hasattr(self, '_metadata'):
+            self._metadata = Metadata(self.doc)
+        return self._metadata
+
+    @property
+    def include_tables(self) -> bool:
+        return self.options.get('tables', True)
+
+    @property
+    def include_images(self) -> bool:
+        return self.options.get('images', True)
+
+    @property
+    def include_styles(self) -> bool:
+        return self.options.get('styles', True)
+
+    def save(self, destination) -> None:
+        """Save the document to a file path or BytesIO object."""
+        if isinstance(destination, str):
+            destination, _ = os.path.splitext(destination)
+            self.doc.save(f'{destination}.docx')
+        elif isinstance(destination, BytesIO):
+            self.doc.save(destination)
+        else:
+            raise TypeError('destination must be a str path or BytesIO object')
 
     def copy_settings_from(self, other):
         """Copy settings from another instance of HtmlToDocx"""
@@ -109,112 +110,148 @@ class HtmlToDocx(HTMLParser):
         tcPr = tc.get_or_add_tcPr()  # Get or add the table cell properties
 
         # Default border properties
-        default_size = 0
-        default_color = "#000000"
-        default_style = "single"
-        borders = {
-            "top": {"size": default_size, "color": default_color, "style": default_style},
-            "right": {"size": default_size, "color": default_color, "style": default_style},
-            "bottom": {"size": default_size, "color": default_color, "style": default_style},
-            "left": {"size": default_size, "color": default_color, "style": default_style},
-        }
-        border_styles = {
-            "none": "none",
-            "hidden": "none",
-            "initial": "none",
-            "solid": "single",
-            "dotted": "dotted",
-            "dashed": "dashed",
-            "double": "double",
-            "inset": "inset",
-            "outset": "outset"
-        }
+        default_size = constants.DEFAULT_BORDER_SIZE
+        default_color = constants.DEFAULT_BORDER_COLOR
+        default_style = constants.DEFAULT_BORDER_STYLE
+        borders = constants.default_borders()
+        border_styles = constants.BORDER_STYLES
+        keywords = constants.BORDER_KEYWORDS
+        border_sides = ("top", "right", "bottom", "left")
+        border_width_pattern = re.compile(r'^[0-9]*\.?[0-9]+(px|pt|cm|in|rem|em|%)$')
 
-        def parse_border_style(value: str):
+        def parse_border_style(value: str) -> str:
             """Parses border styles to match word standart"""
-            return border_styles[value] if value in border_styles.keys() else 'none'
+            return constants.BORDER_STYLES[value] if value in constants.BORDER_STYLES.keys() else 'none'
 
+        def check_unit_keywords(value: str) -> str:
+            """Convert medium, thin, thick keywords to numeric values (px)"""
+            lower_val = value.lower()
+            return keywords.get(lower_val, value)
+
+        @lru_cache(maxsize=None)
         def border_unit_converter(unit_value: str):
             """Convert multiple units to pt that is used on Word table cell border"""
             unit_value = utils.remove_important_from_style(unit_value)
+            unit_value = check_unit_keywords(unit_value)
+
+            # Return default if no value or empty
+            if not unit_value or unit_value == '':
+                return default_size
+
             unit = re.sub(r'[0-9\.]+', '', unit_value)
             value = float(re.sub(r'[a-zA-Z\!\%]+', '', unit_value))  # Allow float values
 
             if unit == 'px':
-                result = int(value * 0.75)  # 1 px = 0.75 pt
+                result = value * 0.75  # 1 px = 0.75 pt
             elif unit == 'cm':
-                result = int(value * 28.35)  # 1 cm = 28.35 pt
+                result = value * 28.35  # 1 cm = 28.35 pt
             elif unit == 'in':
-                result = int(value * 72)  # 1 inch = 72 pt
+                result = value * 72  # 1 inch = 72 pt
             elif unit == 'pt':
-                result = int(value) # default is pt
+                result = value # default is pt
             elif unit == 'rem' or unit == 'em':
-                result = int(value * 12)  # Assuming 1rem/em = 16px, converted to pt
+                result = value * 12  # Assuming 1rem/em = 16px, converted to pt
             elif unit == '%':
-                result = int(MAX_INDENT * (value / 100))
+                result = constants.MAX_INDENT * (value / 100)
             else:
                 return None  # Unsupported units return None
 
             return result
 
         def parse_border_value(value: str):
-            """Parses a border value like '1px solid #000000' or '5px'"""
+            """
+            Parses a border value like:
+            '1px solid #000000', 'solid 1px red', or '#000000 medium dashed' in any order.
+            """
             parts = value.split()
-            size = border_unit_converter(parts[0]) if parts else default_size
-            style = parse_border_style(parts[1]) if len(parts) > 1 else default_style
-            color = utils.parse_color(parts[2], return_hex=True) if len(parts) > 2 else default_color
+
+            # Return all default if there is only 'none' or empty
+            if (len(parts) == 1 and parts[0] == 'none') or (not value or value.strip() == ''):
+                return default_size, default_style, default_color
+
+            size = None
+            style = default_style
+            color = default_color
+
+            for part in parts:
+                clean_part = utils.remove_important_from_style(part).lower()
+
+                # Detect size (units or keywords)
+                if border_width_pattern.match(clean_part) or clean_part in keywords:
+                    size = border_unit_converter(clean_part) or default_size
+                    continue
+
+                # Detect style
+                if clean_part in border_styles:
+                    style = parse_border_style(clean_part)
+                    continue
+
+                # Detect color
+                if utils.is_color(clean_part):
+                    color = utils.parse_color(clean_part, return_hex=True)
+                    continue
+
+            # If only style or color was given without size, we need to apply 1pt size to render it
+            if len(parts) >= 1 and size is None:
+                size = 1.0
+
             return size, style, color
 
-        for style, value in styles.items():
-            if "border" not in style:
+        for css_prop, css_value in styles.items():
+            if not css_prop.startswith("border"):
                 continue
 
-            if style == "border":  # Handle shorthand border
-                border_values = value.split()
-                num_values = len(border_values)
-
-                if num_values == 1:  # "5px"
-                    size, style, color = parse_border_value(value)
-                    for side in borders:
+            # Case 1: 'border' shorthand applies to all sides
+            if css_prop == "border":
+                values = css_value.split()
+                if len(values) in (1, 3):
+                    # Single value or full triple — parse all parts in any order
+                    size, style, color = parse_border_value(css_value)
+                    for side in border_sides:
                         borders[side].update({"size": size, "style": style, "color": color})
-                elif num_values == 2:  # "10px 20px"
-                    borders["top"].update({"size": border_unit_converter(border_values[0])})
-                    borders["bottom"].update({"size": border_unit_converter(border_values[0])})
-                    borders["left"].update({"size": border_unit_converter(border_values[1])})
-                    borders["right"].update({"size": border_unit_converter(border_values[1])})
-                elif num_values == 3:  # "5px solid #000000"
-                    size, style, color = parse_border_value(value)
-                    for side in borders:
-                        borders[side].update({"size": size, "style": style, "color": color})
-                elif num_values == 4:  # "1px 5px 10px 20px"
-                    borders["top"].update({"size": border_unit_converter(border_values[0])})
-                    borders["right"].update({"size": border_unit_converter(border_values[1])})
-                    borders["bottom"].update({"size": border_unit_converter(border_values[2])})
-                    borders["left"].update({"size": border_unit_converter(border_values[3])})
+                elif len(values) == 2:
+                    # Two widths (top/bottom, left/right)
+                    tb_size = border_unit_converter(values[0]) or default_size
+                    lr_size = border_unit_converter(values[1]) or default_size
+                    for side in ("top", "bottom"):
+                        borders[side]["size"] = tb_size
+                    for side in ("left", "right"):
+                        borders[side]["size"] = lr_size
+                elif len(values) == 4:
+                    # Four widths (top, right, bottom, left)
+                    for side, val in zip(border_sides, values):
+                        borders[side]["size"] = border_unit_converter(val) or default_size
 
-            elif style in ("border-width", "border-color", "border-style"):
-                for side in borders:
-                    prop = style.split("-")[-1]
-                    if prop == "width":
-                        borders[side]["size"] = border_unit_converter(value)
-                    elif prop == "color":
-                        borders[side]["color"] = utils.parse_color(value, return_hex=True)
-                    elif prop == "style":
-                        borders[side]["style"] = parse_border_style(value)
+            # Case 2: 'border-width', 'border-color', 'border-style' — apply to all sides
+            elif css_prop in ("border-width", "border-color", "border-style"):
+                prop_type = css_prop.split("-")[1]
+                if prop_type == "width":
+                    size = border_unit_converter(css_value) or default_size
+                    for side in border_sides:
+                        borders[side]["size"] = size
+                elif prop_type == "color":
+                    color = utils.parse_color(css_value, return_hex=True)
+                    for side in border_sides:
+                        borders[side]["color"] = color
+                elif prop_type == "style":
+                    style = parse_border_style(css_value.lower())
+                    for side in border_sides:
+                        borders[side]["style"] = style
 
-            elif re.match(r"^border-(top|right|bottom|left)(-(width|color|style))?$", style):
-                parts = style.split("-")
+            # Case 3: 'border-top', 'border-right-width', etc.
+            else:
+                parts = css_prop.split("-")
                 side = parts[1]
-                prop = parts[2] if len(parts) > 2 else None
-
-                if prop == "width":
-                    borders[side]["size"] = border_unit_converter(value)
-                elif prop == "color":
-                    borders[side]["color"] = utils.parse_color(value, return_hex=True)
-                elif prop == "style":
-                    borders[side]["style"] = parse_border_style(value)
+                prop_type = parts[2] if len(parts) > 2 else None
+                if prop_type == "width":
+                    borders[side]["size"] = border_unit_converter(css_value) or default_size
+                elif prop_type == "color":
+                    borders[side]["color"] = utils.parse_color(css_value, return_hex=True)
+                elif prop_type == "style":
+                    borders[side]["style"] = parse_border_style(css_value.lower())
                 else:
-                    size, style, color = parse_border_value(value)
+                    # Full side shorthand in any order
+                    size, style, color = parse_border_value(css_value)
                     borders[side].update({"size": size, "style": style, "color": color})
 
         # Check if w:tcBorders exists, otherwise create it
@@ -237,34 +274,39 @@ class HtmlToDocx(HTMLParser):
         bookmark_start = OxmlElement('w:bookmarkStart')
         bookmark_start.set(qn('w:id'), str(self.bookmark_id))
         bookmark_start.set(qn('w:name'), bookmark_name)
-        self.paragraph._element.insert(0, bookmark_start)
 
         bookmark_end = OxmlElement('w:bookmarkEnd')
         bookmark_end.set(qn('w:id'), str(self.bookmark_id))
-        self.paragraph._element.append(bookmark_end)
 
+        if not self.paragraph:
+            self.paragraph = self.doc.add_paragraph()
+
+        self.paragraph._element.insert(0, bookmark_start)
+        self.paragraph._element.append(bookmark_end)
         self.bookmark_id += 1
 
     def add_text_align_or_margin_to(self, obj, style):
+        """Styles that can be applied on multiple objects"""
         if 'text-align' in style:
             align = utils.remove_important_from_style(style['text-align'])
 
             if 'center' in align:
                 obj.alignment = WD_ALIGN_PARAGRAPH.CENTER
+            elif 'left' in align:
+                obj.alignment = WD_ALIGN_PARAGRAPH.LEFT
             elif 'right' in align:
                 obj.alignment = WD_ALIGN_PARAGRAPH.RIGHT
             elif 'justify' in align:
                 obj.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY
 
         if 'margin-left' in style and 'margin-right' in style:
-            margin_left = style['margin-left']
-            margin_right = style['margin-right']
-            if 'auto' in margin_left and 'auto' in margin_right:
+            if 'auto' in style['margin-left'] and 'auto' in style['margin-right']:
                 obj.alignment = WD_ALIGN_PARAGRAPH.CENTER
         elif 'margin-left' in style:
             obj.left_indent = utils.unit_converter(style['margin-left'])
 
     def add_styles_to_table_cell(self, styles, doc_cell, cell_row):
+        """Styles that must be applied specifically in a _Cell object"""
         # Set background color
         if 'background-color' in styles:
             self.set_cell_background(doc_cell, styles['background-color'])
@@ -285,11 +327,22 @@ class HtmlToDocx(HTMLParser):
                     for run in paragraph.runs:
                         run.font.color.rgb = color
 
+        # Set vertical align (for individual cells)
+        if 'vertical-align' in styles:
+            align = utils.remove_important_from_style(styles['vertical-align'])
+
+            if 'top' in align:
+                doc_cell.vertical_alignment = WD_ALIGN_VERTICAL.TOP
+            elif 'middle' in align:
+                doc_cell.vertical_alignment = WD_ALIGN_VERTICAL.CENTER
+            elif 'bottom' in align:
+                doc_cell.vertical_alignment = WD_ALIGN_VERTICAL.BOTTOM
+
         # Set borders
         if any('border' in style for style in styles.keys()):
             self.set_cell_borders(doc_cell, styles)
 
-        self.add_text_align_or_margin_to(doc_cell, styles)
+        self.add_text_align_or_margin_to(doc_cell.paragraphs[0], styles)
 
     def add_styles_to_run(self, style):
         if 'font-size' in style:
@@ -311,10 +364,7 @@ class HtmlToDocx(HTMLParser):
             # Little trick to apply background-color to paragraph
             # because `self.run.font.highlight_color`
             # has a very limited amount of colors
-            #
-            # Create XML element
             shd = OxmlElement('w:shd')
-            # Add attributes to the element
             shd.set(qn('w:val'), 'clear')
             shd.set(qn('w:color'), 'auto')
             shd.set(qn('w:fill'), color.lstrip('#'))
@@ -326,21 +376,64 @@ class HtmlToDocx(HTMLParser):
             self.paragraph.paragraph_format.element.pPr.append(shd)
 
     def handle_li(self):
-        # check list stack to determine style and depth
-        list_depth = len(self.tags['list'])
-        if list_depth:
-            list_type = self.tags['list'][-1]
-        else:
-            list_type = 'ul' # assign unordered if no tag
-
-        if list_type == 'ol':
-            list_style = utils.styles['LIST_NUMBER']
-        else:
-            list_style = utils.styles['LIST_BULLET']
+        '''
+            Handle li tags
+            source: https://stackoverflow.com/a/78685353/17274446
+        '''
+        list_depth = len(self.tags['list']) or 1
+        list_type = self.tags['list'][-1] if self.tags['list'] else 'ul'
+        level = min(list_depth, 3)
+        style_key = list_type if level <= 1 else f"{list_type}{level}"
+        list_style = constants.STYLES.get(style_key, 'List Number' if list_type == 'ol' else 'List Bullet')
 
         self.paragraph = self.doc.add_paragraph(style=list_style)
-        self.paragraph.paragraph_format.left_indent = Inches(min(list_depth * LIST_INDENT, MAX_INDENT))
-        self.paragraph.paragraph_format.line_spacing = 1
+        self.in_li = True
+
+        if list_type == "ol":
+            # Use your current_ol_num_id (generated on <ol> open) as key
+            ol_id = self.current_ol_num_id or -1
+
+            if ol_id not in self._list_num_ids:
+                # First time using this <ol> → create a new numId
+
+                style_obj = self.paragraph.style
+                num_id_style = None
+
+                if hasattr(style_obj._element.pPr, 'numPr'):
+                    num_id_style = style_obj._element.pPr.numPr.numId.val
+
+                if num_id_style is not None:
+                    ct_numbering = self.doc.part.numbering_part.numbering_definitions._numbering
+                    ct_num = ct_numbering.num_having_numId(num_id_style)
+                    abstractNumId = ct_num.abstractNumId.val
+
+                    # Add new numId linked to same abstractNumId
+                    ct_num_new = ct_numbering.add_num(abstractNumId)
+                    new_num_id = ct_num_new.numId
+
+                    # Apply startOverride for level 0
+                    lvl_override = ct_num_new.add_lvlOverride(0)
+                    start_override = lvl_override._add_startOverride()
+                    start_override.val = 1
+
+                    # Cache this new numId
+                    self._list_num_ids[ol_id] = new_num_id
+            else:
+                new_num_id = self._list_num_ids[ol_id]
+
+            # Assign this numId to the paragraph
+            pPr = self.paragraph._p.get_or_add_pPr()
+            numPr = OxmlElement('w:numPr')
+
+            numId_elem = OxmlElement('w:numId')
+            numId_elem.set(qn('w:val'), str(new_num_id))
+
+            ilvl = OxmlElement('w:ilvl')
+            ilvl.set(qn('w:val'), str(level - 1))
+
+            numPr.append(ilvl)
+            numPr.append(numId_elem)
+            pPr.append(numPr)
 
     def add_image_to_cell(self, cell, image, width=None, height=None):
         # python-docx doesn't have method yet for adding images to table cells. For now we use this
@@ -367,11 +460,16 @@ class HtmlToDocx(HTMLParser):
         # fetch image
         image = utils.fetch_image_data(src)
 
+        if not self.paragraph:
+            self.paragraph = self.doc.add_paragraph()
+
+        self.run = self.paragraph.add_run()
+
         # add image to doc
         if image:
             try:
                 if isinstance(self.doc, docx.document.Document):
-                    self.doc.add_picture(image, width, height)
+                    self.run.add_picture(image, width, height)
                 else:
                     self.add_image_to_cell(self.doc, image, width, height)
             except FileNotFoundError:
@@ -409,41 +507,74 @@ class HtmlToDocx(HTMLParser):
         """
         table_soup = self.tables[self.table_no]
         rows, cols = self.get_table_dimensions(table_soup)
-        # Available Table Styles
-        # https://python-docx.readthedocs.io/en/latest/user/styles-understanding.html#table-styles-in-default-template
         self.table = self.doc.add_table(rows, cols)
 
         if self.table_style:
+            # Available Table Styles
+            # https://python-docx.readthedocs.io/en/latest/user/styles-understanding.html#table-styles-in-default-template
             try:
                 # Fixed 'style lookup by style_id is deprecated.'
                 # https://stackoverflow.com/a/29567907/17274446
                 self.table_style = ' '.join(re.findall(r'[A-Z][a-z]*|[0-9]', self.table_style))
+                # Available Table Styles
+                # https://python-docx.readthedocs.io/en/latest/user/styles-understanding.html#table-styles-in-default-template
                 self.table.style = self.table_style
             except KeyError as e:
                 raise ValueError(f"Unable to apply style {self.table_style}.") from e
 
         # Reference:
         # https://python-docx.readthedocs.io/en/latest/api/table.html#cell-objects
+        used_cells = [[False] * cols for _ in range(rows)]
         for cell_row, row in enumerate(self.get_table_rows(table_soup)):
-            for cell_col, col in enumerate(self.get_table_columns(row)):
+            col_offset = 0  # Shift index if some columns are occupied
+            for col in self.get_table_columns(row):
+                while used_cells[cell_row][col_offset]:
+                    col_offset += 1
+
+                current_row = cell_row
+                current_col = col_offset
+
                 cell_html = self.get_cell_html(col)
                 if col.name == 'th':
                     cell_html = "<b>%s</b>" % cell_html
 
                 # Get _Cell object from table based on cell_row and cell_col
-                docx_cell = self.table.cell(cell_row, cell_col)
+                docx_cell = self.table.cell(current_row, current_col)
+
+                # Reference:
+                # https://python-docx.readthedocs.io/en/latest/dev/analysis/features/table/cell-merge.html
+                rowspan = int(col.get('rowspan', 1))
+                colspan = int(col.get('colspan', 1))
+
+                if rowspan > 1 or colspan > 1:
+                    docx_cell = docx_cell.merge(
+                        self.table.cell(
+                            current_row + (rowspan - 1),
+                            current_col + (colspan - 1)
+                        )
+                    )
+
+                    # Mark all merged cells as used
+                    for r in range(current_row, current_row + rowspan):
+                        for c in range(current_col, current_col + colspan):
+                            used_cells[r][c] = True
+                else:
+                    used_cells[current_row][current_col] = True
 
                 # Parse cell styles
                 cell_styles = utils.parse_dict_string(col.get('style', ''))
 
                 if 'width' in cell_styles or 'height' in cell_styles:
                     self.table.autofit = False
+                    self.table.allow_autofit = False
 
                 child_parser = HtmlToDocx()
                 child_parser.copy_settings_from(self)
                 child_parser.add_html_to_cell(cell_html, docx_cell)
                 child_parser.add_styles_to_table_cell(cell_styles, docx_cell, self.table.rows[cell_row])
 
+                col_offset += colspan  # Move to the next real column
+
         if 'style' in current_attrs and self.table:
             style = utils.parse_dict_string(current_attrs['style'])
             self.add_text_align_or_margin_to(self.table, style)
@@ -455,10 +586,34 @@ class HtmlToDocx(HTMLParser):
         self.table = None
 
     def handle_div(self, current_attrs):
+        self.paragraph = self.doc.add_paragraph()
+
         # handle page break
         if 'style' in current_attrs and 'page-break-after: always' in current_attrs['style']:
             self.doc.add_page_break()
 
+    def handle_hr(self):
+        # This implementation was taken from:
+        # https://github.com/python-openxml/python-docx/issues/105#issuecomment-62806373
+        self.paragraph = self.doc.add_paragraph()
+        pPr = self.paragraph._p.get_or_add_pPr()
+        pBdr = OxmlElement('w:pBdr')
+        pPr.insert_element_before(
+            pBdr,
+            'w:shd', 'w:tabs', 'w:suppressAutoHyphens', 'w:kinsoku', 'w:wordWrap',
+            'w:overflowPunct', 'w:topLinePunct', 'w:autoSpaceDE', 'w:autoSpaceDN',
+            'w:bidi', 'w:adjustRightInd', 'w:snapToGrid', 'w:spacing', 'w:ind',
+            'w:contextualSpacing', 'w:mirrorIndents', 'w:suppressOverlap', 'w:jc',
+            'w:textDirection', 'w:textAlignment', 'w:textboxTightWrap',
+            'w:outlineLvl', 'w:divId', 'w:cnfStyle', 'w:rPr', 'w:sectPr', 'w:pPrChange'
+        )
+        bottom = OxmlElement('w:bottom')
+        bottom.set(qn('w:val'), 'single')
+        bottom.set(qn('w:sz'), '6')
+        bottom.set(qn('w:space'), '1')
+        bottom.set(qn('w:color'), 'auto')
+        pBdr.append(bottom)
+
     def handle_link(self, href, text, tooltip=None):
         """
         A function that places a hyperlink within a paragraph object.
@@ -468,7 +623,7 @@ class HtmlToDocx(HTMLParser):
             text: The text displayed for the url.
             tooltip: The text displayed when holder link.
         """
-        is_external = href.startswith('http')
+        is_external = href.startswith('http') if href else False
         hyperlink = OxmlElement('w:hyperlink')
 
         if is_external:
@@ -528,7 +683,14 @@ class HtmlToDocx(HTMLParser):
         if tag == 'span':
             self.tags['span'].append(current_attrs)
             return
-        elif tag == 'ol' or tag == 'ul':
+        elif tag in ['ol', 'ul']:
+            if tag == 'ol':
+                # Assign new ID if it's a fresh top-level list
+                self.list_restart_counter += 1
+                self.current_ol_num_id = self.list_restart_counter
+            else:
+                self.current_ol_num_id = None  # unordered list
+
             self.tags['list'].append(tag)
             return # don't apply styles for now
         elif tag == 'br':
@@ -542,33 +704,14 @@ class HtmlToDocx(HTMLParser):
 
         self.tags[tag] = current_attrs
         if tag in ['p', 'pre']:
-            self.paragraph = self.doc.add_paragraph()
+            if not self.in_li:
+                self.paragraph = self.doc.add_paragraph()
 
         elif tag == 'li':
             self.handle_li()
 
         elif tag == 'hr':
-            # This implementation was taken from:
-            # https://github.com/python-openxml/python-docx/issues/105#issuecomment-62806373
-            self.paragraph = self.doc.add_paragraph()
-            pPr = self.paragraph._p.get_or_add_pPr()
-            pBdr = OxmlElement('w:pBdr')
-            pPr.insert_element_before(
-                pBdr,
-                'w:shd', 'w:tabs', 'w:suppressAutoHyphens', 'w:kinsoku', 'w:wordWrap',
-                'w:overflowPunct', 'w:topLinePunct', 'w:autoSpaceDE', 'w:autoSpaceDN',
-                'w:bidi', 'w:adjustRightInd', 'w:snapToGrid', 'w:spacing', 'w:ind',
-                'w:contextualSpacing', 'w:mirrorIndents', 'w:suppressOverlap', 'w:jc',
-                'w:textDirection', 'w:textAlignment', 'w:textboxTightWrap',
-                'w:outlineLvl', 'w:divId', 'w:cnfStyle', 'w:rPr', 'w:sectPr',
-                'w:pPrChange'
-            )
-            bottom = OxmlElement('w:bottom')
-            bottom.set(qn('w:val'), 'single')
-            bottom.set(qn('w:sz'), '6')
-            bottom.set(qn('w:space'), '1')
-            bottom.set(qn('w:color'), 'auto')
-            pBdr.append(bottom)
+            self.handle_hr()
 
         elif re.match('h[1-9]', tag):
             if isinstance(self.doc, docx.document.Document):
@@ -582,8 +725,9 @@ class HtmlToDocx(HTMLParser):
             self.paragraph = self.doc.paragraphs[-1]
 
         elif tag == 'table':
-            self.handle_table(current_attrs)
-            return
+            if self.include_tables:
+                self.handle_table(current_attrs)
+                return
 
         elif tag == 'div':
             self.handle_div(current_attrs)
@@ -620,14 +764,19 @@ class HtmlToDocx(HTMLParser):
             if self.tags['span']:
                 self.tags['span'].pop()
                 return
-        elif tag == 'ol' or tag == 'ul':
+        elif tag in ['ol', 'ul']:
             utils.remove_last_occurence(self.tags['list'], tag)
+            if tag == 'ol':
+                self._list_num_ids.pop(self.current_ol_num_id, None)
+                self.current_ol_num_id = None
             return
         elif tag == 'table':
-            self.table_no += 1
-            self.table = None
-            self.doc = self.document
-            self.paragraph = None
+            if self.include_tables:
+                self.table_no += 1
+                self.table = None
+                self.paragraph = None
+        elif tag == 'li':
+            self.in_li = False
 
         if tag in self.tags:
             self.tags.pop(tag)
@@ -648,27 +797,35 @@ class HtmlToDocx(HTMLParser):
         # There can only be one nested link in a valid html document
         # You cannot have interactive content in an A tag, this includes links
         # https://html.spec.whatwg.org/#interactive-content
-        link = self.tags.get('a')
-        if link:
-            self.handle_link(link.get('href', None), data, link.get('title', None))
-        else:
-            # If there's a link, dont put the data directly in the run
-            self.run = self.paragraph.add_run(data)
-            spans = self.tags['span']
-            for span in spans:
-                if 'style' in span:
-                    style = utils.parse_dict_string(span['style'])
-                    self.add_styles_to_run(style)
+        link = self.tags.get('a', {})
+        href = link.get('href', None)
+        title = link.get('title', None)
 
-            # add font style and name
-            for tag in self.tags:
-                if tag in utils.font_styles:
-                    font_style = utils.font_styles[tag]
-                    setattr(self.run.font, font_style, True)
+        if link and href:
+            self.handle_link(href, data, title)
+            return
 
-                if tag in utils.font_names:
-                    font_name = utils.font_names[tag]
-                    self.run.font.name = font_name
+        # If there's a link, dont put the data directly in the run
+        self.run = self.paragraph.add_run(data)
+
+        for span in self.tags['span']:
+            if 'style' in span:
+                style = utils.parse_dict_string(span['style'])
+                self.add_styles_to_run(style)
+
+        # add font style and name
+        for tag, attrs in self.tags.items():
+            if tag in constants.FONT_STYLES:
+                font_style = constants.FONT_STYLES[tag]
+                setattr(self.run.font, font_style, True)
+
+            if tag in constants.FONT_NAMES:
+                font_name = constants.FONT_NAMES[tag]
+                self.run.font.name = font_name
+
+            if 'style' in attrs and (tag in ['div', 'li', 'p', 'pre'] or re.match(r'h[1-9]', tag)):
+                style = utils.parse_dict_string(attrs['style'])
+                self.add_styles_to_run(style)
 
     def ignore_nested_tables(self, tables_soup):
         """
@@ -702,40 +859,58 @@ class HtmlToDocx(HTMLParser):
         # Table is either empty or has non-direct children between table and tr tags
         # Thus the row dimensions and column dimensions are assumed to be 0
         # A table can have a varying number of columns per row,
-        #     so it is important to find the maximum number of columns in any row
-        if rows:
-            cols = max(len(self.get_table_columns(row)) for row in rows)
-        else:
-            cols = 0
-        return len(rows), cols
+        # so it is important to find the maximum number of columns in any row
+        if not rows:
+            return 0, 0
 
-    def get_tables(self):
+        default_span = 1
+        max_cols = 0
+        max_rows = len(rows)
+
+        for row_idx, row in enumerate(rows):
+            cols = self.get_table_columns(row)
+            # Handle colspan
+            row_col_count = sum(int(col.get('colspan', default_span)) for col in cols)
+            max_cols = max(max_cols, row_col_count)
+
+            # Handle rowspan
+            for col in cols:
+                rowspan = int(col.get('rowspan', default_span))
+                if rowspan > default_span:
+                    max_rows = max(max_rows, row_idx + rowspan)
+
+        return max_rows, max_cols
+
+    def get_tables(self) -> None:
         if not hasattr(self, 'soup'):
-            self.include_tables = False
+            self.options['tables'] = False
             return
-            # find other way to do it, or require this dependency?
+
         self.tables = self.ignore_nested_tables(self.soup.find_all('table'))
         self.table_no = 0
 
-    def run_process(self, html):
+    def run_process(self, html: str) -> None:
         if self.bs and BeautifulSoup:
             self.soup = BeautifulSoup(html, 'html.parser')
+
             html = str(self.soup)
         if self.include_tables:
             self.get_tables()
         self.feed(html)
 
-    def add_html_to_document(self, html, document):
+    def add_html_to_document(self, html: str, document) -> None:
         if not isinstance(html, str):
             raise ValueError(f'First argument needs to be a {str}')
         elif not isinstance(document, docx.document.Document) and not isinstance(document, docx.table._Cell):
             raise ValueError(f'Second argument needs to be a {docx.document.Document}')
+
         self.set_initial_attrs(document)
         self.run_process(html)
 
-    def add_html_to_cell(self, html, cell):
+    def add_html_to_cell(self, html: str, cell: docx.table._Cell) -> None:
         if not isinstance(cell, docx.table._Cell):
             raise ValueError(f'Second argument needs to be a {docx.table._Cell}')
+
         unwanted_paragraph = cell.paragraphs[0]
         utils.delete_paragraph(unwanted_paragraph)
         self.set_initial_attrs(cell)
@@ -745,17 +920,20 @@ class HtmlToDocx(HTMLParser):
         if not self.doc.paragraphs:
             self.doc.add_paragraph('')
 
-    def parse_html_file(self, filename_html, filename_docx=None, encoding='utf-8'):
+    def parse_html_file(self, filename_html: str, filename_docx, encoding: str = 'utf-8') -> None:
         with open(filename_html, 'r', encoding=encoding) as infile:
             html = infile.read()
+
         self.set_initial_attrs()
         self.run_process(html)
+
         if not filename_docx:
             path, filename = os.path.split(filename_html)
             filename_docx = f'{path}/new_docx_file_{filename}'
-        self.doc.save(f'{filename_docx}.docx')
 
-    def parse_html_string(self, html):
+        self.save(filename_docx)
+
+    def parse_html_string(self, html: str) -> docx.document.Document:
         self.set_initial_attrs()
         self.run_process(html)
         return self.doc
